@@ -34,6 +34,7 @@ public:
     size_t maxObs,
     bool hasNas,
     bool linear,
+    bool symmetric,
     double overfitPenalty,
     unsigned int seed
   );
@@ -117,7 +118,10 @@ public:
     std::shared_ptr< arma::Mat<double> > gtotal,
     std::shared_ptr< arma::Mat<double> > stotal,
     bool monotone_splits,
-    monotonic_info monotone_details
+    monotonic_info monotone_details,
+    bool trinary,
+    bool centerSplit,
+    symmetric_info symmetric_details
   );
 
   void selectBestFeature(
@@ -125,6 +129,8 @@ public:
       double &bestSplitValue,
       double &bestSplitLoss,
       int &bestSplitNaDir,
+      std::vector<double> &bestSplitLeftWts,
+      std::vector<double> &bestSplitRightWts,
       arma::Mat<double> &bestSplitGL,
       arma::Mat<double> &bestSplitGR,
       arma::Mat<double> &bestSplitSL,
@@ -137,11 +143,13 @@ public:
       bool splitMiddle,
       size_t maxObs,
       bool linear,
+      bool trinary,
       double overfitPenalty,
       std::shared_ptr< arma::Mat<double> > gtotal,
       std::shared_ptr< arma::Mat<double> > stotal,
       bool monotone_splits,
-      monotonic_info &monotone_details
+      monotonic_info &monotone_details,
+      symmetric_info &symmetric_details
   );
 
   void initializelinear(
