@@ -135,7 +135,8 @@ void multilayerForestry::addForests(size_t ntree) {
       std::move(residualobservationWeights_),
       std::move(monotonicConstraintsRcpp_),
       std::move(groupMembershipRcpp_),
-      false
+      false,
+      0
     );
 
     forestry *residualForest = new forestry(
@@ -162,6 +163,7 @@ void multilayerForestry::addForests(size_t ntree) {
       0, // We aren't giving minimum trees per groups, so set this to zero
       false,  // This is hasNAs being set to false
       _linear,
+      false, // Set symmetric = FALSE
       _overfitPenalty,
       _doubleTree
     );
