@@ -42,7 +42,8 @@ SEXP rcpp_cppDataFrameInterface(
     Rcpp::NumericVector observationWeights,
     Rcpp::NumericVector monotonicConstraints,
     Rcpp::NumericVector groupMemberships,
-    bool monotoneAvg
+    bool monotoneAvg,
+    Rcpp::NumericVector symmetricIndices
 ){
 
   try {
@@ -114,6 +115,11 @@ SEXP rcpp_cppDataFrameInterface(
         )
     );
 
+    std::unique_ptr< std::vector<size_t> > symmetricIndicesRcpp (
+        new std::vector<size_t>(
+            Rcpp::as< std::vector<size_t> >(symmetricIndices)
+        )
+    );
 
     DataFrame* trainingData = new DataFrame(
         std::move(featureDataRcpp),
@@ -129,7 +135,8 @@ SEXP rcpp_cppDataFrameInterface(
         std::move(observationWeightsRcpp),
         std::move(monotonicConstraintsRcpp),
         std::move(groupMembershipsRcpp),
-        (bool) monotoneAvg
+        (bool) monotoneAvg,
+        std::move(symmetricIndicesRcpp)
     );
 
     Rcpp::XPtr<DataFrame> ptr(trainingData, true) ;
@@ -178,10 +185,12 @@ SEXP rcpp_cppBuildInterface(
   Rcpp::NumericVector observationWeights,
   Rcpp::NumericVector monotonicConstraints,
   Rcpp::NumericVector groupMemberships,
+  Rcpp::NumericVector symmetricIndices,
   int minTreesPerGroup,
   bool monotoneAvg,
   bool hasNas,
   bool linear,
+  bool symmetric,
   double overfitPenalty,
   bool doubleTree,
   bool existing_dataframe_flag,
@@ -217,6 +226,7 @@ SEXP rcpp_cppBuildInterface(
         (size_t) minTreesPerGroup,
         hasNas,
         linear,
+        symmetric,
         (double) overfitPenalty,
         doubleTree
       );
@@ -307,6 +317,12 @@ SEXP rcpp_cppBuildInterface(
           )
       );
 
+      std::unique_ptr< std::vector<size_t> > symmetricIndicesRcpp (
+          new std::vector<size_t>(
+              Rcpp::as< std::vector<size_t> >(symmetricIndices)
+          )
+      );
+
       DataFrame* trainingData = new DataFrame(
           std::move(featureDataRcpp),
           std::move(outcomeDataRcpp),
@@ -321,7 +337,8 @@ SEXP rcpp_cppBuildInterface(
           std::move(observationWeightsRcpp),
           std::move(monotoneConstraintsRcpp),
           std::move(groupMembershipsRcpp),
-          (bool) monotoneAvg
+          (bool) monotoneAvg,
+          std::move(symmetricIndicesRcpp)
       );
 
       forestry* testFullForest = new forestry(
@@ -348,6 +365,7 @@ SEXP rcpp_cppBuildInterface(
         (size_t) minTreesPerGroup,
         hasNas,
         linear,
+        symmetric,
         (double) overfitPenalty,
         doubleTree
       );
@@ -1087,8 +1105,10 @@ Rcpp::List rcpp_reconstructree(
   Rcpp::NumericVector monotonicConstraints,
   Rcpp::NumericVector groupMemberships,
   bool monotoneAvg,
+  int symmetricIndex,
   bool hasNas,
   bool linear,
+  Rcpp::NumericVector symmetric,
   double overfitPenalty,
   bool doubleTree
 ){
@@ -1240,6 +1260,12 @@ Rcpp::List rcpp_reconstructree(
       )
   );
 
+  std::unique_ptr< std::vector<size_t> > symmetricIndicesRcpp (
+      new std::vector<size_t>(
+          Rcpp::as< std::vector<size_t> >(groupMemberships)
+      )
+  );
+
   DataFrame* trainingData = new DataFrame(
     std::move(featureDataRcpp),
     std::move(outcomeDataRcpp),
@@ -1254,7 +1280,8 @@ Rcpp::List rcpp_reconstructree(
     std::move(observationWeightsRcpp),
     std::move(monotonicConstraintsRcpp),
     std::move(groupMembershipsRcpp),
-    (bool) monotoneAvg
+    (bool) monotoneAvg,
+    std::move(symmetricIndicesRcpp)
   );
 
   forestry* testFullForest = new forestry(
@@ -1281,6 +1308,7 @@ Rcpp::List rcpp_reconstructree(
     (size_t) minTreesPerGroup,
     (bool) hasNas,
     (bool) linear,
+    (bool) symmetric,
     (double) overfitPenalty,
     doubleTree
   );
@@ -1348,7 +1376,9 @@ Rcpp::List rcpp_reconstruct_forests(
     Rcpp::NumericVector groupMemberships,
     Rcpp::NumericVector gammas,
     bool monotoneAvg,
+    int symmetricIndex,
     bool linear,
+    Rcpp::NumericVector symmetric,
     double overfitPenalty,
     bool doubleTree
 ){
@@ -1536,6 +1566,12 @@ Rcpp::List rcpp_reconstruct_forests(
         )
     );
 
+    std::unique_ptr< std::vector<size_t> > symmetricIndicesRcpp (
+        new std::vector<size_t>(
+            Rcpp::as< std::vector<size_t> >(groupMemberships)
+        )
+    );
+
 
     DataFrame* trainingData = new DataFrame(
       std::move(featureDataRcpp),
@@ -1551,7 +1587,8 @@ Rcpp::List rcpp_reconstruct_forests(
       std::move(observationWeightsRcpp),
       std::move(monotonicConstraintsRcpp),
       std::move(groupMembershipsRcpp),
-      (bool) monotoneAvg
+      (bool) monotoneAvg,
+      std::move(symmetricIndicesRcpp)
     );
 
     // std::cout << "Making a forest \n";
@@ -1581,6 +1618,7 @@ Rcpp::List rcpp_reconstruct_forests(
       (size_t) minTreesPerGroup,
       false,
       (bool) linear,
+      (bool) symmetric,
       (double) overfitPenalty,
       doubleTree
     );
@@ -1679,6 +1717,11 @@ Rcpp::List rcpp_reconstruct_forests(
           Rcpp::as< std::vector<size_t> >(groupMemberships)
       )
   );
+  std::unique_ptr< std::vector<size_t> > symmetricIndicesRcpp (
+      new std::vector<size_t>(
+          Rcpp::as< std::vector<size_t> >(groupMemberships)
+      )
+  );
 
   DataFrame* trainingData = new DataFrame(
     std::move(featureDataRcpp),
@@ -1694,7 +1737,8 @@ Rcpp::List rcpp_reconstruct_forests(
     std::move(observationWeightsRcpp),
     std::move(monotonicConstraintsRcpp),
     std::move(groupMembershipsRcpp),
-    (bool) monotoneAvg
+    (bool) monotoneAvg,
+    std::move(symmetricIndicesRcpp)
   );
 
 
