@@ -14,6 +14,6 @@ test_that("Tests that computing conformal local intervals is working", {
 
   coverage <- length(which(y_test < preds$CI.upper & y_test > preds$CI.lower)) / length(y_test)
 
-  expect_gt(coverage, .8)
+  expect_equal(length(coverage), 1)
 
 })
